#!/usr/bin/env php
<?php

/*
 * LMS version 1.11-git
 *
 *  (C) Copyright 2001-2020 LMS Developers
 *
 *  Please, see the doc/AUTHORS for more information about authors!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 *
 *  $Id$
 */

ini_set('error_reporting', E_ALL & ~E_NOTICE);

$parameters = array(
    'config-file:' => 'C:',
    'quiet' => 'q',
    'help' => 'h',
    'version' => 'v',
    'test' => 't',
    'fakedate:' => 'f:',
    'ignore-send-date' => null,
    'extra-file:' => 'e:',
    'customerid:' => null,
    'division:' => null,
    'customergroups:' => null,
    'customer-status:' => null,
);

$long_to_shorts = array();
foreach ($parameters as $long => $short) {
    $long = str_replace(':', '', $long);
    if (isset($short)) {
        $short = str_replace(':', '', $short);
    }
    $long_to_shorts[$long] = $short;
}

$options = getopt(
    implode(
        '',
        array_filter(
            array_values($parameters),
            function ($value) {
                return isset($value);
            }
        )
    ),
    array_keys($parameters)
);

foreach (array_flip(array_filter($long_to_shorts, function ($value) {
    return isset($value);
})) as $short => $long) {
    if (array_key_exists($short, $options)) {
        $options[$long] = $options[$short];
        unset($options[$short]);
    }
}

if (array_key_exists('version', $options)) {
    print <<<EOF
lms-sendinvoices.php
(C) 2001-2020 LMS Developers

EOF;
    exit(0);
}

if (array_key_exists('help', $options)) {
    print <<<EOF
lms-sendinvoices.php
(C) 2001-2020 LMS Developers

-C, --config-file=/etc/lms/lms.ini      alternate config file (default: /etc/lms/lms.ini);
-h, --help                      print this help and exit;
-t, --test                      print only invoices to send;
-v, --version                   print version info and exit;
-q, --quiet                     suppress any output, except errors;
-f, --fakedate=YYYY/MM/DD       override system date;
                                (can be specified as percentage value);
    --ignore-send-date          send documents which have already been sent earlier;
-e, --extra-file=/tmp/file.pdf  send additional file as attachment
    --customerid=<id>           limit invoices to specifed customer
    --division=<shortname>
                                limit assignments to customers which belong to specified
                                division
    --customergroups=<group1,group2,...>
                                allow to specify customer groups to which notified customers
                                should be assigned
    --customer-status=<status1,status2,...>
                                send invoices of customers with specified status only

EOF;
    exit(0);
}

$quiet = array_key_exists('quiet', $options);
if (!$quiet) {
    print <<<EOF
lms-sendinvoices.php
(C) 2001-2020 LMS Developers

EOF;
}
if (array_key_exists('config-file', $options)) {
    $CONFIG_FILE = $options['config-file'];
} else {
    $CONFIG_FILE = DIRECTORY_SEPARATOR . 'etc' . DIRECTORY_SEPARATOR . 'lms' . DIRECTORY_SEPARATOR . 'lms.ini';
}

if (!$quiet) {
    echo "Using file ".$CONFIG_FILE." as config." . PHP_EOL;
}

if (!is_readable($CONFIG_FILE)) {
    die("Unable to read configuration file [".$CONFIG_FILE."]!" . PHP_EOL);
}

define('CONFIG_FILE', $CONFIG_FILE);

$CONFIG = (array) parse_ini_file($CONFIG_FILE, true);

// Check for configuration vars and set default values
$CONFIG['directories']['sys_dir'] = (!isset($CONFIG['directories']['sys_dir']) ? getcwd() : $CONFIG['directories']['sys_dir']);
$CONFIG['directories']['lib_dir'] = (!isset($CONFIG['directories']['lib_dir']) ? $CONFIG['directories']['sys_dir'] . DIRECTORY_SEPARATOR . 'lib' : $CONFIG['directories']['lib_dir']);
$CONFIG['directories']['doc_dir'] = (!isset($CONFIG['directories']['doc_dir']) ? $CONFIG['directories']['sys_dir'] . DIRECTORY_SEPARATOR . 'documents' : $CONFIG['directories']['doc_dir']);
$CONFIG['directories']['smarty_compile_dir'] = (!isset($CONFIG['directories']['smarty_compile_dir']) ? $CONFIG['directories']['sys_dir'] . DIRECTORY_SEPARATOR . 'templates_c' : $CONFIG['directories']['smarty_compile_dir']);
$CONFIG['directories']['smarty_templates_dir'] = (!isset($CONFIG['directories']['smarty_templates_dir']) ? $CONFIG['directories']['sys_dir'] . DIRECTORY_SEPARATOR . 'templates' : $CONFIG['directories']['smarty_templates_dir']);
$CONFIG['directories']['plugin_dir'] = (!isset($CONFIG['directories']['plugin_dir']) ? $CONFIG['directories']['sys_dir'] . DIRECTORY_SEPARATOR . 'plugins' : $CONFIG['directories']['plugin_dir']);
$CONFIG['directories']['plugins_dir'] = $CONFIG['directories']['plugin_dir'];

define('SYS_DIR', $CONFIG['directories']['sys_dir']);
define('LIB_DIR', $CONFIG['directories']['lib_dir']);
define('DOC_DIR', $CONFIG['directories']['doc_dir']);
define('SMARTY_COMPILE_DIR', $CONFIG['directories']['smarty_compile_dir']);
define('SMARTY_TEMPLATES_DIR', $CONFIG['directories']['smarty_templates_dir']);
define('PLUGIN_DIR', $CONFIG['directories']['plugin_dir']);
define('PLUGINS_DIR', $CONFIG['directories']['plugin_dir']);

define('K_TCPDF_EXTERNAL_CONFIG', true);

// Load autoloader
$composer_autoload_path = SYS_DIR . DIRECTORY_SEPARATOR . 'vendor' . DIRECTORY_SEPARATOR . 'autoload.php';
if (file_exists($composer_autoload_path)) {
    require_once $composer_autoload_path;
} else {
    die("Composer autoload not found. Run 'composer install' command from LMS directory and try again. More informations at https://getcomposer.org/" . PHP_EOL);
}

// Init database

$DB = null;

try {
    $DB = LMSDB::getInstance();
} catch (Exception $ex) {
    trigger_error($ex->getMessage(), E_USER_WARNING);
    // can't working without database
    die("Fatal error: cannot connect to database!" . PHP_EOL);
}

$invoice_filetype = ConfigHelper::getConfig('invoices.type', ''); // TODO
$dnote_filetype = ConfigHelper::getConfig('notes.type', '');
$disabled_protection = ConfigHelper::getConfig('invoices.disable_protection', '');

if ($invoice_filetype!="pdf" || $dnote_filetype!="pdf")
    die("You can send documents in PDF format only. Please change invoices.type and notes.type parameters to pdf value." . PHP_EOL);

if ($disabled_protection!=1 )
    die("PDF documents cannot be protected against modification. Please change invoices.disable_protection parameter to value of 1" . PHP_EOL);
// Initialize templates engine (must be before locale settings)
$SMARTY = new LMSSmarty;

// test for proper version of Smarty

if (defined('Smarty::SMARTY_VERSION')) {
    $ver_chunks = preg_split('/[- ]/', preg_replace('/^smarty-/i', '', Smarty::SMARTY_VERSION), -1, PREG_SPLIT_NO_EMPTY);
} else {
    $ver_chunks = null;
}
if (count($ver_chunks) < 1 || version_compare('3.1', $ver_chunks[0]) > 0) {
    die('Wrong version of Smarty engine! We support only Smarty-3.x greater than 3.1.' . PHP_EOL);
}

define('SMARTY_VERSION', $ver_chunks[0]);

// add LMS's custom plugins directory
$SMARTY->addPluginsDir(LIB_DIR . DIRECTORY_SEPARATOR . 'SmartyPlugins');


// Include required files (including sequence is important)

require_once(LIB_DIR . DIRECTORY_SEPARATOR . 'common.php');
require_once(LIB_DIR . DIRECTORY_SEPARATOR . 'language.php');
include_once(LIB_DIR . DIRECTORY_SEPARATOR . 'definitions.php');

$SYSLOG = SYSLOG::getInstance();

// Initialize Session, Auth and LMS classes
$AUTH = null;
$LMS = new LMS($DB, $AUTH, $SYSLOG);

$plugin_manager = new LMSPluginManager();
$LMS->setPluginManager($plugin_manager);

$divisionid = isset($options['division']) ? $LMS->getDivisionIdByShortName($options['division']) : null;
if (!empty($divisionid)) {
    ConfigHelper::setFilter($divisionid);
}

// Set some template and layout variables

$SMARTY->setTemplateDir(null);
$custom_templates_dir = ConfigHelper::getConfig('phpui.custom_templates_dir');
if (!empty($custom_templates_dir) && file_exists(SMARTY_TEMPLATES_DIR . DIRECTORY_SEPARATOR . $custom_templates_dir)
    && !is_file(SMARTY_TEMPLATES_DIR . DIRECTORY_SEPARATOR . $custom_templates_dir)) {
    $SMARTY->AddTemplateDir(SMARTY_TEMPLATES_DIR . DIRECTORY_SEPARATOR . $custom_templates_dir);
}
$SMARTY->AddTemplateDir(
    array(
        SMARTY_TEMPLATES_DIR . DIRECTORY_SEPARATOR . 'default',
        SMARTY_TEMPLATES_DIR,
    )
);
$SMARTY->setCompileDir(SMARTY_COMPILE_DIR);

$SMARTY->assignByRef('layout', $layout);

$customergroups = ConfigHelper::getConfig('postivo.customergroups', '', true); //TODO

$allowed_customer_status = Utils::determineAllowedCustomerStatus(
    isset($options['customer-status'])
        ? $options['customer-status']
        : ConfigHelper::getConfig('postivo.allowed_customer_status', ''),
    -1
);

if (empty($allowed_customer_status)) {
    $customer_status_condition = '';
} else {
    $customer_status_condition = ' AND c.status IN (' . implode(',', $allowed_customer_status) . ')';
}

$extrafile = (array_key_exists('extra-file', $options) ? $options['extra-file'] : null);
if ($extrafile && !is_readable($extrafile)) {
    die("Unable to read additional file [$extrafile]!" . PHP_EOL);
}

$fakedate = isset($options['fakedate']) ? $options['fakedate'] : null;
$customerid = isset($options['customerid']) && intval($options['customerid']) ? $options['customerid'] : null;

function localtime2($fakedate)
{
    if (!empty($fakedate)) {
        $date = explode("/", $fakedate);
        return mktime(0, 0, 0, intval($date[1]), intval($date[2]), intval($date[0]));
    } else {
        return time();
    }
}

$timeoffset = date('Z');
$currtime = localtime2($fakedate) + $timeoffset;
$daystart = (intval($currtime / 86400) * 86400) - $timeoffset;
$dayend = $daystart + 86399;

// prepare customergroups in sql query
if (isset($options['customergroups'])) {
    $customergroups = $options['customergroups'];
}
if (!empty($customergroups)) {
    $ORs = preg_split("/([\s]+|[\s]*,[\s]*)/", mb_strtoupper($customergroups), -1, PREG_SPLIT_NO_EMPTY);
    $customergroup_ORs = array();
    foreach ($ORs as $OR) {
        $ANDs = preg_split("/([\s]*\+[\s]*)/", $OR, -1, PREG_SPLIT_NO_EMPTY);
        $customergroup_ANDs_regular = array();
        $customergroup_ANDs_inversed = array();
        foreach ($ANDs as $AND) {
            if (strpos($AND, '!') === false) {
                $customergroup_ANDs_regular[] = $AND;
            } else {
                $customergroup_ANDs_inversed[] = substr($AND, 1);
            }
        }
        $customergroup_ORs[] = '('
            . (empty($customergroup_ANDs_regular) ? '1 = 1' : "EXISTS (SELECT COUNT(*) FROM customergroups
            JOIN customerassignments ON customerassignments.customergroupid = customergroups.id
            WHERE customerassignments.customerid = c.id
            AND UPPER(customergroups.name) IN ('" . implode("', '", $customergroup_ANDs_regular) . "')
            HAVING COUNT(*) = " . count($customergroup_ANDs_regular) . ')')
            . (empty($customergroup_ANDs_inversed) ? '' : " AND NOT EXISTS (SELECT COUNT(*) FROM customergroups
            JOIN customerassignments ON customerassignments.customergroupid = customergroups.id
            WHERE customerassignments.customerid = c.id
            AND UPPER(customergroups.name) IN ('" . implode("', '", $customergroup_ANDs_inversed) . "')
            HAVING COUNT(*) > 0)")
            . ')';
    }
    $customergroups = ' AND (' . implode(' OR ', $customergroup_ORs) . ')';
}

$test = array_key_exists('test', $options);
if ($test) {
    echo "WARNING! You are using test mode." . PHP_EOL;
}

$plugin_manager->executeHook('smarty_initialized', $SMARTY);

$args = array(DOC_INVOICE, DOC_INVOICE_PRO, DOC_CNOTE, DOC_DNOTE);

$ignore_send_date = isset($options['ignore-send-date']) || ConfigHelper::checkConfig('postivo.ignore_send_date');

$query = "SELECT d.id, d.number, d.cdate, d.customerid, d.type AS doctype, n.template, c.name, c.lastname, a.name AS recipient_name, ca.address_id, a.city, a.postoffice, a.street, a.zip, a.house, a.flat 
		FROM documents d
		LEFT JOIN customeraddressview c ON c.id = d.customerid
        JOIN customer_addresses ca ON ca.customer_id = c.id 
        LEFT JOIN addresses a ON ca.address_id =a.id
		LEFT JOIN numberplans n ON n.id = d.numberplanid 
		WHERE " . ($customerid ? 'c.id = ' . $customerid : '1 = 1')
            . $customer_status_condition
            . ($divisionid ? ' AND c.divisionid = ' . $divisionid : '')
            . " AND ca.type=(SELECT MIN(s2.type) FROM customer_addresses s2 WHERE ca.customer_id = s2.customer_id) AND c.deleted = 0 AND d.cancelled = 0 AND d.type IN (?, ?, ?, ?)
			 AND d.cdate >= $daystart AND d.cdate <= $dayend"
            . ($customergroups ?: '')
        . " ORDER BY d.number";
$docs = $DB->GetAll($query, $args);

if (!empty($docs)) {
    $docs_to_send = array();
    foreach ($docs as $doc) {
        if ($ignore_send_date || empty($doc['senddate'])) {
            $docs_to_send[] = $doc;
        }
    }
    if (empty($docs_to_send)) {
        die;
    }
    $docs = $docs_to_send;
    $which = DOC_ENTITY_ORIGINAL;

    $sLogin = ConfigHelper::getConfig('postivo.login', '');
    $sApiPass = ConfigHelper::getConfig('postivo.password', '');
    $iConfigId = ConfigHelper::getConfig('postivo.config_id', '');
    $iSenderId = ConfigHelper::getConfig('postivo.sender_id', '');
    $bSandbox = ConfigHelper::getConfig('postivo.sandbox', 0);

    $arDispatchOptions = array();
    if ($iSenderId)
        $arDispatchOptions[] = array("name" => "sender_id", "value" => $iSenderId);
    if ($iConfigId)
        $arDispatchOptions[] = array("name" => "config_id", "value" => $iConfigId);

    try{
        if ($bSandbox)
            $client = new SoapClient("https://api.postivo.pl/webservices-sandbox.php?wsdl");
        else
            $client = new SoapClient("https://api.postivo.pl/webservices.php?wsdl");

        foreach ($docs as $doc) {
            $document = $LMS->GetTradeDocument($doc);


            $invoice_number = (!empty($doc['template']) ? $doc['template'] : '%N/LMS/%Y');
            $invoice_number = docnumber(array(
                'number' => $doc['number'],
                'template' => $invoice_number,
                'cdate' => $doc['cdate'] + date('Z'),
                'customerid' => $doc['customerid'],
            ));

            if (!$quiet)
                echo "Sending document no. " . $invoice_number . PHP_EOL;

            if (!$test) {
                try{
                    $arDocFiles = array();
                    $arDocFiles[] = array("file_stream" => base64_encode($document['data']), "file_name" => "doc.pdf");
                    if ($extrafile && is_readable($extrafile))
                        $arDocFiles[] = array("file_stream" => base64_encode(file_get_contents($extrafile)), "file_name" => "extrafile.pdf");

                    $arRecipients = array(
                        array("id" => 0 ,"source" => "inline", "recipient_name" => ($doc['recipient_name'] ? $doc['recipient_name'] : $doc['name']." ".$doc['lastname']),
                            "recipient_address_prefix" => "", "recipient_address" => ($doc['postoffice'] && $doc['city'] && $doc['postoffice']<>$doc['city'] ? $doc['city'].", " : "") . $doc['street'],
                            "recipient_home_number" => ($doc['house'] ? $doc['house'] : " "), "recipient_flat_number" => $doc['flat'],
                            "recipient_post_code" => $doc['zip'], "recipient_city" => ($doc['postoffice'] ? $doc['postoffice'] : $doc['city'])
                        )
                    );
                    $output = $client->dispatch($sLogin, $sApiPass, 1, $arDocFiles, $arRecipients, $arDispatchOptions);
                    if (!$quiet) {
                        if ($output->result == "OK"){
                            if ($output->shipments[0]->status=="OK")
                                echo "Document no. " . $invoice_number ." was succesfully sent to Postivo.pl [Shipment ID: ".$output->shipments[0]->id."]". PHP_EOL;
                            else
                                echo "There was a problem with sending document no. " . $invoice_number ." to Postivo.pl [".$output->shipments[0]->status_description."]". PHP_EOL;
                        }
                        else
                            echo "There was a problem with sending document no. " . $invoice_number ." to Postivo.pl [". $output->result_description."]". PHP_EOL;
                    }
                }
                catch(Exception $oE)
                {
                    if (!$quiet)
                        echo "There was an unknown error with: ".$oE->getMessage();
                }

            }
        }
    }
    catch(Exception $oE){
        if (!$quiet)
            echo "There was an unknown error with: ".$oE->getMessage();
    }
}